package rice.pastry.socket;

import java.io.IOException;
import java.net.*;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;

import rice.environment.Environment;
import rice.environment.logging.Logger;
import rice.environment.params.Parameters;
import rice.pastry.socket.messaging.IPAddressRequestMessage;
import rice.pastry.socket.messaging.IPAddressResponseMessage;

public class StubNATHandler implements NATHandler {
  public static final int ALWAYS = 1;
  public static final int PREFIX_MATCH = 2;
  public static final int NEVER = 3;
  public static final int OVERWRITE = 1;
  public static final int USE_DIFFERENT_PORT = 2;
  public static final int FAIL = 3;

  Environment environment;
  Logger logger;
  InetSocketAddress localAddress;
  InetSocketAddress pAddress;
  
  public StubNATHandler(Environment env, InetSocketAddress localAddress, InetSocketAddress proxyAddress) {
    this.environment = env;
    this.localAddress = localAddress;
    this.pAddress = proxyAddress;
    this.logger = env.getLogManager().getLogger(this.getClass(),"");
  }
  
  /**
   * Method which can be used to test the connectivity contstrains of the local
   * node. This (optional) method is designed to be called by applications to
   * ensure that the local node is able to connect through the network - checks
   * can be done to check TCP/UDP connectivity, firewall setup, etc...
   * 
   * If the method works, then nothing should be done and the method should
   * return. If an error condition is detected, an exception should be thrown.
   */
  public static InetSocketAddress verifyConnection(int timeout,
      InetSocketAddress local, InetSocketAddress[] existing, Environment env,
      Logger logger) throws IOException {
    if (logger.level <= Logger.INFO)
      logger.log("Verifying connection of local node " + local + " using "
          + existing[0] + " and " + existing.length + " more");
    DatagramSocket socket = null;

    try {
      socket = new DatagramSocket(local);
      socket.setSoTimeout(timeout);

      for (int i = 0; i < existing.length; i++) {
        byte[] buf = PingManager
            .addHeader(SourceRoute
                .build(new EpochInetSocketAddress(existing[i])),
                new IPAddressRequestMessage(env.getTimeSource()
                    .currentTimeMillis()), new EpochInetSocketAddress(local),
                env, logger);
        DatagramPacket send = new DatagramPacket(buf, buf.length, existing[i]);
        socket.send(send);
      }

      DatagramPacket receive = new DatagramPacket(new byte[10000], 10000);
      socket.receive(receive);

      byte[] data = new byte[receive.getLength() - 38];
      System.arraycopy(receive.getData(), 38, data, 0, data.length);

      return ((IPAddressResponseMessage) PingManager.deserialize(data, env,
          null, logger)).getAddress();
    } finally {
      if (socket != null)
        socket.close();
    }
  }
  
  /**
   * Internal method which returns a random subset of the address to ping in parallel
   *
   * @param array The list of all hosts
   * @param num The number of hosts to return
   * @return A subset of the specified length
   */
  protected InetSocketAddress[] randomSubset(InetSocketAddress[] array, int num) {
    InetSocketAddress[] result = new InetSocketAddress[num];
    
    for (int i=0; i<result.length; i++)
      result[i] = array[environment.getRandomSource().nextInt(array.length)];
    
    return result;
  }

  /**
   * @return true if ip address matches firewall prefix
   */
  public boolean localAddressIsProbablyNatted() {
    String ip = localAddress.getAddress().getHostAddress();
    String nattedNetworkPrefixes = environment.getParameters().getString("nat_network_prefixes");
    
    String[] nattedNetworkPrefix = nattedNetworkPrefixes.split(";");
    for (int i = 0; i < nattedNetworkPrefix.length; i++) {
      if (ip.startsWith(nattedNetworkPrefix[i])) {
        return true;
      }
    }
    return false;
  }

  protected int getFireWallPolicyVariable(String key) {
    String val = environment.getParameters().getString(key);
    if (val.equalsIgnoreCase("prefix")) return PREFIX_MATCH;
    if (val.equalsIgnoreCase("change")) return USE_DIFFERENT_PORT;
    if (val.equalsIgnoreCase("never")) return NEVER;
    if (val.equalsIgnoreCase("overwrite")) return OVERWRITE;
    if (val.equalsIgnoreCase("always")) return ALWAYS;
    if (val.equalsIgnoreCase("fail")) return FAIL;
    throw new RuntimeException("Unknown value "+val+" for "+key);
  }

  public synchronized InetSocketAddress findExternalIPwithoutUPnP() {
    InetSocketAddress natAddress = null;
    Parameters parameters = environment.getParameters();
    InetSocketAddress[] addresses;

    switch (getFireWallPolicyVariable("nat_search_policy")) {
    case NEVER:
      return null;
    case PREFIX_MATCH:
      if (!localAddressIsProbablyNatted())
        return null;
    case ALWAYS:
      // fall through
    default:  
      // fall through
    }
    
    
    try {
      addresses = parameters.getInetSocketAddressArray("pastry_proxy_connectivity_hosts");
    } catch (UnknownHostException e1) {
      if (logger.level <= Logger.INFO) logger.logException("problem parsing option pastry_proxy_connectivity_hosts: ",e1);
      return null;
    }
    
    try {
      natAddress = verifyConnection(parameters.getInt("pastry_proxy_connectivity_timeout")/4,
                                                            localAddress,
                                                            randomSubset(addresses, 5), environment, logger);
    } catch (IOException e) {} 
  
    if (natAddress == null) 
      try {
        natAddress = verifyConnection(parameters.getInt("pastry_proxy_connectivity_timeout")/2,
                                                              localAddress,
                                                              randomSubset(addresses, 5), environment, logger);
      } catch (IOException e) {}
      
    if (natAddress == null) 
      try {
        natAddress = verifyConnection(parameters.getInt("pastry_proxy_connectivity_timeout"),
                                                              localAddress,
                                                              randomSubset(addresses, 5), environment, logger);
      } catch (IOException e) {}

    return natAddress;
  }
  
  public boolean usingNAT() {
    return !pAddress.getAddress().equals(localAddress.getAddress());
  }

  public InetSocketAddress externalAddress() throws IOException {
    if (pAddress == null)
      pAddress = findExternalIPwithoutUPnP();
    if (pAddress == null)
      pAddress = localAddress;
    return pAddress;
  }

  public boolean test() {
    // TODO Auto-generated method stub
    return true;
  }

}
